@@ -8,6 +8,7 @@ See Also
 --------
 vulkpy.vkarray : Core Module
 vulkpy.random : Random Module
+vulkpy.nn : Neural Network Module
 vulkpy.util : Utility Module
 
 
@@ -23,5 +24,6 @@ Examples
 >>> print(c)
 [4., 5., 6.]
 """
-from .vkarray import GPU, Array
+from .vkarray import GPU, U32Array, Shape, Array, zeros
 from . import random
+from . import nn
